@@ -1,4 +1,4 @@
-/* Copyright 1995-2004,2006-2015,2017-2019
+/* Copyright 1995-2004,2006-2015,2017-2020,2022
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -28,10 +28,7 @@
 
 #include <stdio.h>
 #include <fcntl.h>
-
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif
 #include <unistd.h>
 #ifdef HAVE_IO_H
 #include <io.h>
@@ -208,6 +205,9 @@ scm_i_mode_to_open_flags (SCM mode, int *is_binary, const char *FUNC_NAME)
 	  flags |= O_BINARY;
 #endif
 	  break;
+        case 'e':
+          flags |= O_CLOEXEC;
+          break;
 	case '0':  /* unbuffered: handled later.  */
 	case 'l':  /* line buffered: handled during output.  */
 	  break;
@@ -368,6 +368,9 @@ SCM_DEFINE (scm_i_open_file, "open-file", 2, 0, 1,
 	    "@item +\n"
 	    "Open the port for both input and output.  E.g., @code{r+}: open\n"
 	    "an existing file for both input and output.\n"
+            "@item e\n"
+            "Mark the underlying file descriptor as close-on-exec, as per the\n"
+            "@code{O_CLOEXEC} flag.\n"
 	    "@item 0\n"
 	    "Create an \"unbuffered\" port.  In this case input and output\n"
 	    "operations are passed directly to the underlying port\n"
@@ -452,6 +455,7 @@ scm_i_fdes_to_port (int fdes, long mode_bits, SCM name, unsigned options)
                                                   "file port");
   fp->fdes = fdes;
   fp->options = options;
+  fp->revealed = 0;
 
   port = scm_c_make_port (scm_file_port_type, mode_bits, (scm_t_bits)fp);
   
